"""Copyright 2010 Adam Bark

This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import math

def gravity(m_planet, m_craft, distance):
    """Calculate the force exherted by the planet on the craft"""
    # G=6.673e-11 (N m**2)/kg - Gravitational constant
    # GMm/r**2
    return (6.673e-11 * m_planet * m_craft) / distance**2

def air_temperature(velocity, fluid_density):
    """Calculate the pressure induced temperature of the atmosphere"""
    # R=8.314472 J/(mol K) - gas constant
    # 1/2 _rho v**2 / R
    return (0.5 * fluid_density * velocity**2) / 8.314472

def magnitude(x, y):
        """Calculate the magnitude of a vector given x, y"""
        return math.sqrt(x**2 + y**2)


class Vector(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def dot(self, vec2):
        return self.x * vec2.x + self.y + vec2.y

    def magnitude(self):
        return math.sqrt(self.x**2 + self.y**2)

    def magnitude2(self):
        return self.x**2 + self.y**2

    def normalise(self):
        mag = self.magnitude()
        return Vector(self.x/mag, self.y/mag)

    def __add__(self, vec2):
        return Vector(self.x + vec2.x, self.y + vec2.y)

    def __div__(self, scalar):
        return Vector(self.x / scalar, self.y / scalar)

    def __mul__(self, scalar):
        if type(scalar) == Vector:
            print self.x, self.y
            raise TypeError
        return Vector(self.x * scalar, self.y * scalar)

    def __neg__(self):
        return Vector(-self.x, -self.y)

    def __pow__(self, exponent, z=None):
        v = Vector(0,0)
        for e in range(exponent):
            v += Vector(self.x, self.y)
        return v

    def __radd__(self, vec2):
        return Vector(self.x + vec2.x, self.y + vec2.y)

    def __rmul__(self, scalar):
        return Vector(self.x * scalar, self.y * scalar)

    def __repr__(self):
        return "Vector(%f, %f)" % (self.x, self.y)


class Crash(Exception):
    pass

class Explosion(Exception):
    pass

class Landed(Exception):
    pass


class Planet(object):
    atm_molar_mass = 17.0
    def __init__(self, mass, radius, surface_air_density):
        self.mass = mass
        self.radius = radius
        self.sa_density = surface_air_density

    def get_density(self, radius):
        #print radius, (radius - self.radius)**2
        if abs(radius) <= (self.radius+1):
            raise Crash(radius)
        else:
            return self.sa_density / ((radius - self.radius)+1)**2

    def get_pressure(self, radius, velocity):
            pressure = 0.5 * self.get_density(radius) * velocity**2
            if pressure > Craft.max_pressure:
                print "pressure", pressure
                raise Explosion(pressure)
            else:
                return pressure

class Craft(object):
    max_thrust = 1e6
    max_temperature = 1500
    #max_force = 1e10
    max_pressure = 3e6
    acceleration = 0
    thrust = 0
    specific_heat = 900.0 # J/(kG K)
    area = 1.0 # Cross section
    def __init__(self, mass, size, velocity, direction, position, temperature):
        self.mass = mass
        self.size = size
        self.velocity = velocity
        self.direction = direction
        self.position = position
        self.temperature = temperature
        self.fuel = 4e6
        self.lift = 0

    def apply_force(self, gravity, atmospheric, dt):
        lift = self.position.normalise() * atmospheric.magnitude() * self.lift
        self.fuel -= self.thrust * dt
        if self.fuel <= 0:
            #print "no fuel", self.fuel
            self.fuel = 0
            self.force = gravity + atmospheric + lift
        else:
            #print force, self.thrust * self.direction, self.direction
            self.force = gravity + atmospheric + lift + \
                self.thrust * self.direction
        self.acceleration = self.force * self.mass**-1
        self.velocity += self.acceleration * dt
        self.position += self.velocity * dt

    def radiate(self, dt, T_surroundings=0):
        # E = e o A T**4 dt
        # e = emissivity, o = Stefan's constant, A = area, T = temperature
        delta_T = self.temperature**4 - T_surroundings**4
        return 0.5 * 5.6703e-8 * self.size * delta_T * dt

    def apply_energy(self, energy):
        self.temperature += energy / (self.mass * self.specific_heat)
        if self.temperature > self.max_temperature:
            raise Explosion(self.temperature)

class Simulation(object):
    def __init__(self, craft, planet):
        self.craft = craft
        self.planet = planet
        self.atmos_radius = math.sqrt(self.planet.sa_density/1000) + \
            self.planet.radius

    def energy_change(self, distance):
        density = self.planet.get_density(self.craft.position.magnitude())
        # *1000 as gram moles but kg/m**3
        moles = density / (self.planet.atm_molar_mass*1000) * \
              self.craft.area * distance
        air_temp = air_temperature(
            self.craft.velocity.magnitude(), density)
        loss = 0.5*moles * 8.314472 * self.craft.temperature
        if self.craft.temperature > 1500:
            print self.craft.temperature, air_temp, density, moles, loss,
            self.craft.velocity.magnitude()
        return 0.25*moles * 8.314472 * air_temp - loss

    def step_force(self, dt, turbulence=None):
        distance = self.craft.position.magnitude()
        gravity_force = gravity(self.planet.mass, self.craft.mass, distance) \
                      * -self.craft.position.normalise()
        try:
            atmospheric_force = self.planet.get_pressure(
                distance, self.craft.velocity.magnitude()) * self.craft.size * \
                              -self.craft.velocity.normalise()
        except Crash as crash:
            #print "velocity", self.craft.velocity.dot(-self.craft.position.normalise())
            if abs(self.craft.velocity.dot(
                -self.craft.position.normalise())) < 100:
                raise Landed
            else:
                raise crash
        #if atmospheric_force.magnitude2():
        #    print gravity_force
        #    print atmospheric_force
        self.craft.apply_force(gravity_force, atmospheric_force, dt)
        return self.craft.position, self.craft.velocity

    def step_all(self, dt, turbulence=None): # Maybe assign turbulence to planet?
        # Might need a switch though to differentiate between model
        # and actual run.
        self.step_force(dt, turbulence)
        radiation = self.craft.radiate(2.7, dt)
        self.craft.apply_energy(self.energy_change(
            self.craft.velocity.magnitude()*dt) - radiation)
        return (self.craft.temperature, self.craft.force, self.craft.velocity,
                self.craft.position, self.craft.direction)

    def step(self, dt):
        distance = self.craft.position.magnitude()
        gravity_force = gravity(self.planet.mass, self.craft.mass, distance) \
                      * -self.craft.position.normalise()
        try:
            if self.craft.velocity.magnitude() > 300:
                atmospheric_force = self.planet.get_pressure(
                    distance, self.craft.velocity.magnitude()) * \
                                  self.craft.size * \
                                  -self.craft.velocity.normalise()
                #radiation = self.craft.radiate(2.7, dt)
                self.craft.apply_energy(self.energy_change(
                    self.craft.velocity.magnitude()*dt))# - radiation)
            else:
                atmospheric_force = self.planet.get_density(distance) * \
                                  self.craft.velocity.magnitude()**2 * 0.1 \
                                  * self.craft.size *\
                                  -self.craft.velocity.normalise()
                print atmospheric_force
                print self.craft.position.magnitude(),
                self.craft.velocity.magnitude()
        except Crash as crash:
            if abs(self.craft.velocity.dot(
                -self.craft.position.normalise())) < 100:
                raise Landed
            else:
                raise crash
        self.craft.apply_force(gravity_force, atmospheric_force, dt)
        return self.craft.position, self.craft.velocity

def test():
    import csv, sys
    craft = Craft(1000, 1.0, Vector(500.0, 0.0),
                  Vector(1,0), Vector(0, 4.2e6), 200)
    planet = Planet(1e24, 4e6, 101e3)
    simul = Simulation(craft, planet)
    writer = csv.writer(open(sys.argv[1], "wb"))
    writer.writerow(["Temp", "Force", "V X", "V Y", "P X", "P Y"])
    for i in xrange(200000):
        t, f, v, p, d = simul.step_all(0.167)
        writer.writerow([t, f.magnitude(), v.x, v.y, p.x, p.y])


if __name__ == "__main__":
    import cProfile
    cProfile.run('test()', 'physics.prof')
    import pstats
    p = pstats.Stats('physics.prof')
    p.strip_dirs()
    p.sort_stats('cumulative')
    p.print_stats(20)